
#import <math.h>

#import "GLDroneShip.h"
#import "spriteSystem.h"
#import "game.h"
#import "markedSprites.h"
#import "SocialNetworkingManager.h"
#import "AchievementsManager.h"
#import "StatsManager.h"


static int gNumDroneShips;
static int gDroneLevel;
static int gNumDroneLevelsPlayed;
static int gNumDronesDestroyed;
static int gNumDronesKilledThisLife;
static int gNumRandDrones;


@implementation GLDroneShip


//------------------------------------------------------------------------
// drone ship specific init
//------------------------------------------------------------------------
- (GLSprite *)init:(int)depth
{
	// do base class init
	[super init:depth];
	
	objectType = OBJECT_TYPE_SHIP_DRONE;
	
	collideSize = 0.045f;
	
	// drone is not a physics controlled ship
	bUsePhysics = false;
	
	maximumSpeed = DRONE_MAX_SPEED;
	curSpeed = maximumSpeed;
	speedTarget = maximumSpeed;
	acceleration = DRONE_ACCEL;	
	turnSpeed = DRONE_ROT_SPEED;
	velocity.x = velocity.y = 0.0f;

	health = DRONE_HEALTH;
	
	shotgunProb = 0.0f;
	missileProb = 0.0f;
	bulletTimer = DRONE_BULLET_TIME;
	bulletFireTime = DRONE_BULLET_TIME;
	numBulletsLeftToFire = 0;
	
	shieldDisplayRatio = 0.0f;

	userID = -100;	// all drones use the same user ID, so their bullets don't collide with each other
	
	shieldStrength = 0.0f;
	
	exploScale = 5.0f;
	
	collideTimer = DRONE_COLLIDE_CHECK_TIME;
	flyAwayTimer = 0.0f;
	
	sideVelocity.x = sideVelocity.y = 0.0f;
		
	// add me to the list of drones
	gNumDroneShips++;
	
	return (self);
}


//------------------------------------------------------------------------
// rotate to point at a location
//------------------------------------------------------------------------
- (void) aimAtPoint:(float)x andY:(float)y
{
	rotationTarget = Math_RotationFromVector(posX - x, y - posY);
}


//------------------------------------------------------------------------
// adjust a ship's health by a damage amount
//------------------------------------------------------------------------
- (void) adjustHealth:(GLfloat)damage byID:(int)hitUserID
{
	if (shieldStrength > 0.0f)
	{	// this ship has shields
		shieldStrength -= damage;
		if (shieldStrength <= 0.0f)
		{	// took down the shields
			damage = 0.0;
			shieldDownTimer = shieldDownTime * MAX_FRAMERATE;
		}
		else
		{
			damage = 0.0f;
			
			// set the shields to flash
			shieldDisplayRatio = 1.0f;
		}
	}
	
	health -= damage;		// damage is only done to sub-parts
	hitByUserID = hitUserID;
}


//------------------------------------------------------------------------
// check basic collision (simple distance check)
//------------------------------------------------------------------------
- (bool) checkCollide:(float)x andY:(float)y dam:(float)damageAmt hitBy:(int)hitByUser shieldFlag:(bool*)pbShielded
{
	float l_collideSize;
	if (shieldStrength > 0.0f)
	{	
		l_collideSize = collideSize * scale;
	}
	else
	{
		l_collideSize = droneCollideSize * scale;
	}
	
	float distX = x - posX;
	float distY = y - posY;
	
	if (distX < 0) distX = -distX;
	if (distY < 0) distY = -distY;
	
	// 1st pass collision, simple box
	if ((distX <= l_collideSize) && (distY <= l_collideSize))
	{	// might have collided... do detailed collision
		float distSqr = (distX*distX) + (distY*distY);
		if (distSqr <= (l_collideSize * l_collideSize))
		{	// hit it!
			if (pbShielded != nil)
			{
				if (shieldStrength > 0.0f)
				{
					*pbShielded = true;
				}
				else
				{
					*pbShielded = false;
				}
			}
			
			[self adjustHealth:damageAmt byID:hitByUser];
			
			return(true);
		}
	}
	
	return(false);
}


//-----------------------------------------------------------------
// check collision against a circle
//
// RETURNS:	penetration distance (negative if no collide)
//-----------------------------------------------------------------
- (float) checkCollideWithCircle:(float)x andY:(float)y andRad:(float)radius dam:(float)damageAmt hitBy:(int)hitByUser doPush:(bool)bDoPush
{
	float distX = x - posX;
	float distY = y - posY;
	
	// 1st pass, distance collide with ship ship/shield size
	float collideDist = radius + (collideSize * scale);
	
	float distSqr = (distX*distX) + (distY*distY);
	if (distSqr >= (collideDist * collideDist))
	{	// no collision possible
		return -1.0f;
	}		
	
	float penetration = collideDist - sqrt(distSqr);
	
	[self adjustHealth:damageAmt byID:hitByUser];
	
	if ((hitByUser == gPlayerShipMsgID) && (health > 0.0f) && (bDoPush == true))
	{	// circle collide with player ID... thus a missile... add push to the drone if still alive
		float dist = sqrt(distSqr);
		float forceX = -(distX / dist) * MISSILE_RECOIL_FORCE;
		float forceY = -(distY / dist) * MISSILE_RECOIL_FORCE;

		CGPoint addVel;
		Physics_ApplyForceSimple(shipMass, forceX, forceY, &addVel.x, &addVel.y);
		
		velocity.x += addVel.x;
		velocity.y += addVel.y;
	}
	
	return penetration;
}


//------------------------------------------------------------------------
// check this drone's health, and react to death if necessary
//------------------------------------------------------------------------
- (bool) checkHealth {
	if (bDoDamage == true)
	{
		if (health <= 0.0f)
		{	// this object ran out of health... kill it with an explosion
			
			// adjust explode scale with a bit of randomness
			float baseBoomScale = scale * exploScale;
			float boomScale = baseBoomScale - (Rand_BetweenVals(0.0f, 0.2f) * baseBoomScale);
			
			GLAnim *pExplo = Explosion_Launch(EXPLODE_TYPE_LARGEWAVE, posX, posY, boomScale, sortDepth+1);
			[pExplo setVelocity:velocity.x andY:velocity.y andDamp:0.99f];
			
			// play big explosion sound
			Sound_Launch(SOUNDFX_EXPLODE, 1.0f);
			
			gNumDroneShips--;
			gNumDronesDestroyed++;
			gNumDronesKilledThisLife++;
			gGameScore += droneScore;
			
			// check drones killed this life achievements
			if (gNumDronesKilledThisLife == 10)
			{
				[SocialNetworkingManager unlockAchievement:_10_DRONE_STREAK];
			}
			else if (gNumDronesKilledThisLife == 25)
			{
				[SocialNetworkingManager unlockAchievement:_25_DRONE_STREAK];
			}
			else if (gNumDronesKilledThisLife == 40)
			{
				[SocialNetworkingManager unlockAchievement:_40_DRONE_STREAK];
			}
			else if (gNumDronesKilledThisLife == 60)
			{
				[SocialNetworkingManager unlockAchievement:_60_DRONE_STREAK];
			}
			else if (gNumDronesKilledThisLife == 100)
			{
				[SocialNetworkingManager unlockAchievement:_100_DRONE_STREAK];
			}
			
			// store this kill, and check for more achievements
			[StatsManager increaseStat:kTotalOfflineKills byValue:1];
			[StatsManager storeDatabase];
			
			int totalKills = [StatsManager getStat:kTotalOfflineKills];
			if (totalKills == 100)
			{
				[SocialNetworkingManager unlockAchievement:DILIGENT_TRAINER];
			}
			else if (totalKills == 1000)
			{
				[SocialNetworkingManager unlockAchievement:DRONE_HUNTER];
			}
			else if (totalKills == 10000)
			{
				[SocialNetworkingManager unlockAchievement:TECHNOPHOBE];
			}
            
			if (droneType == DRONE_TYPE_BOSS)
			{	// adjust boss drone count and check achievements
				[StatsManager increaseStat:kTotalBossKills byValue:1];
				totalKills = [StatsManager getStat:kTotalBossKills];
				if (totalKills == 10)
				{
					[SocialNetworkingManager unlockAchievement:_10_BOSS_KILLS];
				}
				else if (totalKills == 20)
				{
					[SocialNetworkingManager unlockAchievement:_20_BOSS_KILLS];
				}
				else if (totalKills == 100)
				{
					[SocialNetworkingManager unlockAchievement:_50_BOSS_KILLS];
				}
			}
			
			// check probability of launching a powerup
			bool bLaunchPowerup;
			if (powerupProbability >= 1.0f)
			{	// always launch a powerup on death
				bLaunchPowerup = true;
			}
			else
			{	// use random probability to determine if this drone type launches a powerup
				float prob = Rand_BetweenVals(0.0f, 1.0f);
				if (prob < powerupProbability)
				{
					bLaunchPowerup = true;
				}
				else
				{
					bLaunchPowerup = false;
				}
			}
			
			if (bLaunchPowerup == true)
			{	// launch a powerup
				int powerupType = -1;
				
				if (gpPlayerSprite->bMissingParts == true)
				{	// player is missing parts... 50% chance of a repair powerup
					if (rand() & 0x1)
					{
						powerupType = POWERUPTYPE_REPAIR;
					}
				}
				
				int randNum;
				
				if (powerupType < 0)
				{
					randNum = rand() & 0x07;	// set type... bit mask out just the low 3 bits of the random value,
                    // thus randNum will be 0..7, for a total of 8 possible values
					if (randNum <= 1)
					{	// 2 in 8 chance of shield powerup (randNum is 0 or 1)
						powerupType = POWERUPTYPE_SHIELD;
					}
					else if (randNum <= 3)
					{	// 2 in 8 chance of ammo powerup (randNum is 2 or 3)
						powerupType = POWERUPTYPE_AMMO;
					}
					else if (randNum <= 5)
					{	// 2 in 8 chance of rapid fire powerup (randNum is 4 or 5)
						powerupType = POWERUPTYPE_RAPIDFIRE;
					}
					else if (randNum == 6)
					{	// 1 in 8 chance of invulnerability powerup (randNum is 6)
						powerupType = POWERUPTYPE_INVULNERABILITY;
					}
					else
					{	// 1 in 8 chance of lives powerup (randNum is 7)
						powerupType = POWERUPTYPE_LIVES;
					}
				}
                
				// set rotation (and thus drift direction)
				randNum = rand();
				float powerupRot = ((float)(randNum % 360000) / 1000.0f);
                
				// set rotation speed
				randNum = rand();
				float powerupRotSpeed = ((float)(randNum % 5000) / 1000.0f) - 2.5f;
                
				// set speed
				randNum = rand();
				float powerupSpeed = ((float)(randNum % 100) / 10000.0f);
                
				GLPowerup_LaunchPowerup(powerupType, posX, posY, sortDepth, 4.0f, powerupRot, powerupRotSpeed, powerupSpeed, -1);
			}
			
			if (droneType == DRONE_TYPE_BOSS)
			{	// if a boss drone, always launch a lives powerup in addition to the previous powerup
				
				// set rotation (and thus drift direction)
				int randNum = rand();
				float powerupRot = ((float)(randNum % 360000) / 1000.0f);
				
				// set rotation speed
				randNum = rand();
				float powerupRotSpeed = ((float)(randNum % 5000) / 1000.0f) - 2.5f;
				
				// set speed
				randNum = rand();
				float powerupSpeed = ((float)(randNum % 100) / 10000.0f);
				
				GLPowerup_LaunchPowerup(POWERUPTYPE_LIVES, posX, posY, sortDepth, 4.0f, powerupRot, powerupRotSpeed, powerupSpeed, -1);
			}
			
			return true;
		}
	}
    
    return false;
}


//------------------------------------------------------------------------
// update the drone
//------------------------------------------------------------------------
- (bool) updateSprite
{
    if (self checkHealth) {
        return true;
    }
	
	// aim at player ship
	if (gpPlayerSprite->bActive == true)
	{
		[self aimAtPoint:gpPlayerSprite->posX andY:gpPlayerSprite->posY];
	}
	
	// adjust ship rotation towards target rotation
	if (rotation != rotationTarget)
	{
		float turnDist = rotationTarget - rotation;
		
		if (turnDist > 0.0f)
		{
			if ((turnDist < turnSpeed) || (turnDist > (360.0f-turnSpeed)))
			{
				rotation = rotationTarget;
			}
			else if (turnDist > 180.0f)
			{
				rotation -= turnSpeed * gFrameTime;
			}
			else
			{
				rotation += turnSpeed * gFrameTime;
			}
		}
		else
		{
			if ((turnDist > -turnSpeed) || (turnDist < (turnSpeed-360.0f)))
			{
				rotation = rotationTarget;
			}
			else if (turnDist < -180.0f)
			{
				rotation += turnSpeed * gFrameTime;
			}
			else
			{
				rotation -= turnSpeed * gFrameTime;
			}
		}
		
		if (rotation < 0.0f)
		{
			rotation += 360.0;
		}
		else if (rotation >= 360.0f)
		{
			rotation -= 360.0f;
		}
	}

	// adjust velocity
	velocity.x += acceleration * -cos(rotation * DEG_TO_RAD) * gFrameTime;
	velocity.y += acceleration * -sin(rotation * DEG_TO_RAD) * gFrameTime;
	
	curSpeed = sqrt((velocity.x * velocity.x) + (velocity.y * velocity.y));
	if (curSpeed > maximumSpeed)
	{	// cap speed at max
		velocity.x = (velocity.x / curSpeed) * maximumSpeed;
		velocity.y = (velocity.y / curSpeed) * maximumSpeed;
		curSpeed = maximumSpeed;
	}
	
	[self adjustPos:(velocity.x * gFrameTime) andY:(velocity.y * gFrameTime)];
	
	// dampen velocity
	velocity.x *= DRONE_LINEAR_DAMPING;
	velocity.y *= DRONE_LINEAR_DAMPING;

	// adjust side velocity based on distance to other drones
	GLSprite *pCloseDrone = MarkedSprites_ClosestDrone(MARKEDSPRITE_TYPE_ENEMY, self);
	if (pCloseDrone != nil)
	{	// close enough to another drone to add sideways thrust
		
		// get thrust angle away from the drone
		float distX = posX - pCloseDrone->posX;
		float distY = posY - pCloseDrone->posY;
		float droneRot = Math_RotationFromVector(distX, distY);
		float thrustRot;
		
		// is rot to drone clockwise or counterclockwise?
		float relDroneRot = droneRot + rotation;
		if (relDroneRot > 360.0f)
		{
			relDroneRot -= 360.0f;
		}
		
		if (relDroneRot <= 180.0f)
		{
			thrustRot = rotation + 90.0f;
		}
		else
		{
			thrustRot = rotation - 90.0f;
		}
		
		sideVelocity.x += acceleration * -cos(thrustRot * DEG_TO_RAD) * gFrameTime;
		sideVelocity.y += acceleration * -sin(thrustRot * DEG_TO_RAD) * gFrameTime;
		
		curSpeed = sqrt((sideVelocity.x * sideVelocity.x) + (sideVelocity.y * sideVelocity.y));
		if (curSpeed > maximumSpeed)
		{	// cap speed at max
			sideVelocity.x = (sideVelocity.x / curSpeed) * maximumSpeed;
			sideVelocity.y = (sideVelocity.y / curSpeed) * maximumSpeed;
		}
		
		[self adjustPos:(sideVelocity.x * gFrameTime) andY:(sideVelocity.y * gFrameTime)];
		
		// dampen velocity
		sideVelocity.x *= DRONE_LINEAR_DAMPING;
		sideVelocity.y *= DRONE_LINEAR_DAMPING;
	}		

	// set screen zoom based on distance from player ship
	float dX = posX - gScreenWorldPos.x;
	float dY = posY - gScreenWorldPos.y;
	distFromPlayer = sqrt((dX*dX) + (dY*dY));
	
	// set camera distance
	Camera_UpdateTargetDist(distFromPlayer);
	
	// set on/off screen
	if (distFromPlayer > OFFSCREEN_DIST)
	{
		bOffScreen = true;
	}
	else
	{
		bOffScreen = false;
	}
	
	if (spriteID == SPRITETEX_DRONE_SUICIDE)
	{	// NOTE: suicide drones don't fire weapons, so do other code
		if ((distFromPlayer <= suicideProximity) && (gpPlayerSprite->bActive == true))
		{	// suicide drone, and its close to the player... detenate
			float penetration;
			MarkedSprites_CollideCircle(MARKEDSPRITE_TYPE_PLAYER, posX, posY, suicideExploRadius, suicideExploDamage, userID, 
										OBJECT_TYPE_SHIP_DRONE, true, &penetration);
			
			health = 0.0f;
		}
	}
	else
	{	// update firing of weapons
		bulletTimer -= gFrameTime;
		
		if (bulletTimer < 0.0f)
		{
			if (numBulletsLeftToFire > 0)
			{	// still have multiple simple bullets left to fire
				numBulletsLeftToFire--;
				if (numBulletsLeftToFire <= 0)
				{
					bulletTimer += bulletFireTime;
				}
				
				CGPoint droneDir;
				droneDir.x = -cos(rotation * DEG_TO_RAD);
				droneDir.y = -sin(rotation * DEG_TO_RAD);
				
				float bulletPosX, bulletPosY;
				bulletPosX = DRONE_BULLET_OFFSET * droneDir.x;
				bulletPosY = DRONE_BULLET_OFFSET * droneDir.y;
				
				bulletPosX += posX;
				bulletPosY += posY;
				
                Weapon_Fire(WEAPON_TYPE_SIMPLE_BULLET, userID, bulletPosX, bulletPosY, rotation, droneDir.x*curSpeed, droneDir.y*curSpeed, false);
                Sound_Launch_Dist(SOUNDFX_BULLET, 1.0f, distFromPlayer);
			}
			else
			{
				bulletTimer += bulletFireTime;
				
				CGPoint droneDir;
				droneDir.x = -cos(rotation * DEG_TO_RAD);
				droneDir.y = -sin(rotation * DEG_TO_RAD);
				
				float bulletPosX, bulletPosY;
				bulletPosX = DRONE_BULLET_OFFSET * droneDir.x;
				bulletPosY = DRONE_BULLET_OFFSET * droneDir.y;
				
				bulletPosX += posX;
				bulletPosY += posY;
				
				int weaponType;
				float bLimited;

                // set weapon difficulty
                if (droneDifficulty < 3)
                {	// lower level drones use limited special weapons
                    bLimited = true;
                }
                else
                {
                    bLimited = false;
                }
                
                // calc type of weapon
                if (shotgunProb >= 1.0f)
                {	// always fire shotgun
                    weaponType = WEAPON_TYPE_SHOTGUN;
                }
                else if (missileProb >= 1.0f)
                {	// always fire missile
                    weaponType = WEAPON_TYPE_MISSILE;
                }
                else if ((shotgunProb <= 0.0f) && (missileProb <= 0.0f))
                {	// always fire basic gun
                    weaponType = WEAPON_TYPE_SIMPLE_BULLET;
                }
                else
                {	// grab weapon type from random probability
                    float randVal = Rand_BetweenVals(0.0f, 1.0f);
                    if (randVal < shotgunProb)
                    {
                        weaponType = WEAPON_TYPE_SHOTGUN;
                    }
                    else if (randVal < (shotgunProb + missileProb))
                    {
                        weaponType = WEAPON_TYPE_MISSILE;
                    }
                    else
                    {
                        weaponType = WEAPON_TYPE_SIMPLE_BULLET;
                    }
                }
				
				Weapon_Fire(weaponType, userID, bulletPosX, bulletPosY, rotation, droneDir.x*curSpeed, droneDir.y*curSpeed, bLimited);
				
				if ((weaponType == WEAPON_TYPE_SIMPLE_BULLET) && (numBulletsToFire > 1))
				{	// need to launch several more simple bullets after this one
					numBulletsLeftToFire = numBulletsToFire - 1;
					
                    bulletTimer = DRONE_MULTIBULLET_TIME;
				}
				
				if (weaponType == WEAPON_TYPE_SHOTGUN)
				{	// apply a recoil force to the drone
					CGPoint forceVect;
					forceVect.x = SHOTGUN_RECOIL_FORCE * cos(rotation * DEG_TO_RAD);
					forceVect.y = SHOTGUN_RECOIL_FORCE * sin(rotation * DEG_TO_RAD);
					
					CGPoint linAccel;
					Physics_ApplyForceSimple(shipMass, forceVect.x, forceVect.y, &linAccel.x, &linAccel.y);
					velocity.x += linAccel.x;
					velocity.y += linAccel.y;
				}
				
				// launch a sound, distance attenuated....
				switch (weaponType)
				{
					case WEAPON_TYPE_SIMPLE_BULLET:
						Sound_Launch_Dist(SOUNDFX_BULLET, 1.0f, distFromPlayer);
						break;
					case WEAPON_TYPE_SHOTGUN:
						Sound_Launch_Dist(SOUNDFX_SHOTGUN, 1.0f, distFromPlayer);
						break;
					case WEAPON_TYPE_MISSILE:
						Sound_Launch_Dist(SOUNDFX_MISSILE, 1.0f, distFromPlayer);
						break;
					default:
						break;
				}
			}				
		}
	}
	
	// adjust shield
	if (shieldDownTimer > 0.0f)
	{	// shield is down
		if (shieldDownTime > 0.0f)
		{	// don't time to shield back up if down time = 0
			shieldDownTimer -= gFrameTime;
			if (shieldDownTimer <= 0.0f)
			{	// put back shield at limited strength
				shieldStrength = maxShieldStrength * 0.25;
			}
		}
	}
	else
	{	// slowly recharge shield
		if ((shieldRechargeRate > 0.0f) && (shieldStrength < maxShieldStrength))
		{
			shieldStrength += shieldRechargeRate * gFrameTime;
			if (shieldStrength > maxShieldStrength)
			{
				shieldStrength = maxShieldStrength;
			}
		}
	}

	return false;
}


//-------------------------------------------------
// draw the current frame of this drone
//-------------------------------------------------
- (void) drawSprite
{
	if (bOffScreen == true)
	{	// don't draw
		return;
	}
	
	spriteTexture = SpriteDefinition_SetSpriteDraw(spriteID, 0);
	
	[super drawSprite];
	
	// display the shield if necessary
	if (shieldDisplayRatio > 0.0f)
	{
		spriteTexture = SpriteDefinition_SetSpriteDraw(SPRITETEX_SHIELD, 0);
		
		float droneScale = scale;
		scale = droneCollideSize * droneScale * 40.0f;
		
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glColor4f(1.0f, 1.0f, 1.0f, shieldDisplayRatio);
		[super drawSprite];
		
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
		
		shieldDisplayRatio -= gFrameTime * (2.0f / MAX_FRAMERATE);
		scale = droneScale;

		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	}
}


//-----------------------------------------------------------------------
// launch a drone ship
//
// IN:	droneType = type of drone
//		droneDiff = difficulty setting for this drone type
//		posX/Y = world position to launch at
//		sortDepth = display depth
//		rotation = rotation to lunch at
//-----------------------------------------------------------------------
+ (GLDroneShip *) GLDroneShip_LaunchWithType:(int) droneType difficulty:(int)droneDiff x:(float)posX y:(float)posY sort:(int)sortDepth rot:(float)rotation
{
	if ((droneType < 0) || (droneType >= NUM_DRONE_TYPES))
	{	// illegal drone type
		NSLog(@"error - bad drone type on launch! (type %d)", droneType);
		return nil;
	}
	
	SpriteDisplayData *pSprite;
	
	pSprite = Sprite_NewDisplaySprite(sortDepth);
	if (pSprite != nil)
	{
		// set drone settings based on type
		DroneTypeData *pDroneTypeList = gDroneTypes[droneType];
		DroneTypeData *pType = &pDroneTypeList[droneDiff];
        
		GLDroneShip *pShip = (GLDroneShip *)[GLDroneShip createSprite:sortDepth];
		pSprite->pSprite = pShip;
		[pShip setScale:pType->scale];
		[pShip setPos:posX andY:posY];
		[pShip setRotation:rotation];
		
		// mark this sprite as an enemy
		MarkedSprites_AddMarkedSprite(pShip, MARKEDSPRITE_TYPE_ENEMY);
        
		pShip->droneType = droneType;
		pShip->droneDifficulty = droneDiff;
		
		pShip->spriteID = pType->spriteID;
        
		pShip->health = pType->health;
		
		pShip->rotationSpeed = pType->rotSpeed;
		pShip->maximumSpeed = pType->maxSpeed + Rand_BetweenVals(0.0f, DRONE_MAX_SPEED_VARIATION);
		pShip->speedTarget = pType->maxSpeed;
		pShip->curSpeed = Rand_BetweenVals(0.0f, pShip->maximumSpeed);
		pShip->acceleration = pType->accel + Rand_BetweenVals(0.0f, DRONE_ACCEL_VARIATION);
		
		pShip->shotgunProb = pType->shotgunProbability;
		pShip->missileProb = pType->missileProbability;
		pShip->bulletFireTime = (float)pType->bulletTime;
		pShip->bulletTimer = (float)(rand() % pType->bulletTime);
		pShip->numBulletsToFire = pType->numBulletsToFire;
		
		pShip->collideSize = pType->shieldCollideSize;
		pShip->shieldStrength = pType->maxShield;
		pShip->maxShieldStrength = pType->maxShield;
		pShip->shieldRechargeRate = pType->shieldRechargeRate;
		pShip->shieldDownTime = pType->shieldDownTime;
		
		pShip->suicideProximity = pType->exploProximity;
		pShip->suicideExploRadius = pType->exploRadius;
		pShip->suicideExploDamage = pType->exploDamage;
		
		pShip->droneScore = pType->score;
		pShip->powerupProbability = pType->powerupProbability;
        
		// set collide size
		CGPoint spriteSize;
		SpriteDefinition_GetSize(SPRITETEX_DRONE_LASER, &spriteSize.x, &spriteSize.y);
		if (spriteSize.x >= spriteSize.y)
		{
			pShip->droneCollideSize = spriteSize.x;
		}
		else
		{
			pShip->droneCollideSize = spriteSize.y;
		}
		
		if (droneType == DRONE_TYPE_BOSS)
		{	// set boss mass
			pShip->shipMass = DRONE_MASS_BOSS;
		}
		else
		{
			pShip->shipMass = DRONE_MASS;
		}
        
		return pShip;
	}
	else
	{
		NSLog(@"--- drone error... no sprites left to allocate");
	}
	
	return(nil);
}


//-------------------------------------------------------------------------
// launch a drone with random position/etc.
//-------------------------------------------------------------------------
+ (void) GLDroneShip_CreateDroneWithType:(int)droneType baseDiff:(int)baseDroneDiff
{
	float posX = (float)((rand() % 572) - 288);
	float posY = (float)((rand() % 1024) - 512);
	
	float dist = sqrt((posX*posX) + (posY*posY));
	posX = posX / dist;
	posY = posY / dist;
	
	// adjust to screen coords, not pixels
#ifdef NOVARIFT_IPAD
	posX *= 3.0f;
	posY *= 3.0f;
#else
	posX *= 2.0f;
	posY *= 2.0f;
#endif
	
	posX += gScreenWorldPos.x;
	posY += gScreenWorldPos.y;
	
	// set drone's difficulty based on wave entry's setting and game difficulty
	int droneDiff;
	if (gCurGameDifficulty == 0)
	{	// easy difficulty, decrease drone diff by one
		droneDiff = baseDroneDiff - 1;
	}
	else if (gCurGameDifficulty == 1)
	{	// normal difficulty, use current setting
		droneDiff = baseDroneDiff;
	}
	else
	{	// hard difficulty, increase drone diff by one
		droneDiff = baseDroneDiff + 1;
	}
	
	GLDroneShip_Launch(droneType, droneDiff, posX, posY, 20, -Math_RotationFromVector(-posX, posY));
}


//-------------------------------------------------------------------------
// run launching of drone ships
//-------------------------------------------------------------------------
+ (void) GLDroneShip_RunLaunching
{
	if (gNumDroneShips <= 0)
	{	// no drones left
        
		bool bDoRandDrones = false;
		
		// go to next game level
		gNumDroneLevelsPlayed++;
		
		if (gNumDroneLevelsPlayed > 0)
		{	// save the game after each wave (other than the first), just in case of a crash....
			Save_SendToFile();
		}
        
#ifdef NOVARIFT_LITE
		// playing lite version
		if (gNumDroneLevelsPlayed >= LITE_DRONE_WAVE_LIMIT)
		{	// exit single player after playing a fixed number of waves
			gMenuNextState = MENU_STATE_SINGLEPLAYER_ENDSCREEN;
			return;
		}
#endif
		
		gDroneLevel++;
		if (gDroneLevel > NUM_DRONE_GAME_LEVELS)
		{
			gDroneLevel = NUM_DRONE_GAME_LEVELS;
		}
		
		if ((gDroneLevel >= NUM_DRONE_GAME_LEVELS) && (gCurGameDifficulty >= 2))
		{	// past max end of set drone waves... start doing randomly created drone waves
			bDoRandDrones = true;
		}
		else
		{	// go to next set drone wave
			if (gDroneLevel >= NUM_DRONE_GAME_LEVELS)
			{	// repeat previous 2 boss wave sets (12 levels)
				gDroneLevel = NUM_DRONE_GAME_LEVELS-12;
				
				if (gCurGameDifficulty < 2)
				{	// increse difficulty if not at max
					gCurGameDifficulty++;
				}
			}
		}
		
		// check for achievements based on # of waves completed
		if (gNumDroneLevelsPlayed == 11)
		{
			[SocialNetworkingManager unlockAchievement:_10_WAVES];
		}
		else if (gNumDroneLevelsPlayed == 21)
		{
			[SocialNetworkingManager unlockAchievement:_20_WAVES];
		}
		else if (gNumDroneLevelsPlayed == 41)
		{
			[SocialNetworkingManager unlockAchievement:_40_WAVES];
		}
		else if (gNumDroneLevelsPlayed == 61)
		{
			[SocialNetworkingManager unlockAchievement:_60_WAVES];
		}
		else if (gNumDroneLevelsPlayed == 81)
		{
			[SocialNetworkingManager unlockAchievement:_80_WAVES];
		}
		
		// check boss level achievements...
		if ((gNumDroneLevelsPlayed > 1) && ((gNumDroneLevelsPlayed % 6) == 1))
		{	// we just finished a boss level
			if ((bDoRandDrones == true) && (gNumRandDrones < 20))
			{	// increase the number of random drones each boss wave
				gNumRandDrones++;
			}
			
			int bossLevel = gNumDroneLevelsPlayed / 6;
			
			//NSLog(@"Achievement: (possible)... single player boss level completed: boss %d at difficulty %d, num deaths = %d, num lives = %d",
			//	  bossLevel, gOfflineGameDifficulty, gNumPlayerDeaths, gNumPlayerShips);
			
			if (bossLevel == 1)
			{	// finished 1st boss wave
				//NSLog(@"Achievement: Getting Started - complete 1st drone boss fight - 2");
				[AchievementsManager achieve:GETTING_STARTED];
				
				if (gNumPlayerDeaths <= 0)
				{	// finished 1st boss using 1st ship
					//NSLog(@"Achievement: Still Kicking - complete 1st boss fight without being killed - 3");
					[SocialNetworkingManager unlockAchievement:STILL_KICKING];
				}
			}
		}
        
		if (bDoRandDrones == false)
		{	// launch current set wave
			DroneSetData *pDroneList = gDroneGameLevels[gDroneLevel];
			while (pDroneList->numDrones > 0)
			{
				int i;
				for (i=0; i<pDroneList->numDrones; i++)
				{
					GLDroneShip_CreateDrone(pDroneList->droneType, pDroneList->droneDifficulty);
				}
                
				pDroneList++;
			}
		}
		else
		{	// launch a randomly created wave of drones
			int i = 0;
			int droneTypeRange;
            
			if ((gNumDroneLevelsPlayed % 6) == 0)
			{	// launch a boss wave
				// always launch at least 4 boss drones
				int j;
				for (j=0; j<4; j++)
				{
					GLDroneShip_CreateDrone(DRONE_TYPE_BOSS, 3);
					i += 2;
				}
                
				droneTypeRange = NUM_DRONE_TYPES;	// boss waves can also launch more boss drones
			}
			else
			{
				droneTypeRange = NUM_DRONE_TYPES-1;	// non-boss wave, don't launch boss drones
			}
			
			while (i < gNumRandDrones)
			{
				int droneType = rand() % droneTypeRange;
				
				GLDroneShip_CreateDrone(droneType, 3);
				
				if (droneType == DRONE_TYPE_BOSS)
				{
					i += 2;
				}
				else
				{
					i++;
				}
			}
		}
		
		// display notification message for this wave
		char tmpText[32];
		
		sprintf(tmpText, "WAVE %d", gNumDroneLevelsPlayed);
		GLFontNotify_LaunchText(COLOR_GREEN, tmpText);
        
		if ((gNumDroneLevelsPlayed % 6) == 0)
		{	// boss wave... add extra message
			sprintf(tmpText, "Warning! Boss Drones!");
			GLFontNotify_LaunchText(COLOR_GREEN, tmpText);
		}
		
	}
}


@end
