
#import <math.h>

#import "GLNetShip.h"
#import "spriteSystem.h"
#import "game.h"
#import "markedSprites.h"
#import "SocialNetworkingManager.h"
#import "AchievementsManager.h"
#import "StatsManager.h"


@implementation GLNetShip


//------------------------------------------------------------------------
// net ship specific init
//------------------------------------------------------------------------
- (GLSprite *)init:(int)depth
{
	// do base class init
	[super init:depth];
	
	objectType = OBJECT_TYPE_SHIP_NET;
	
	collideSize = 0.25f;
		
	// net ships don't deal with damage locally
	bDoDamage = false;
	
	// net controlled ships are physics controlled
	bUsePhysics = true;
	
	bKillMe = false;
	bActive = true;
	
	pNameText = nil;
	pChatIcon = nil;
	bChatting = false;
	
	bNewPos = false;
	velocity.x = 0.0f;
	velocity.y = 0.0f;
	rotationSpeed = 0.0f;
	
	msgPingTime = -1.0f;
	
	shieldStrength = 10.0f;
	
	exploScale = 15.0f;
	
	bChatting = false;
	
	return (self);
}


//------------------------------------------------------------------------
// set this ship to be permanently removed
//------------------------------------------------------------------------
- (void) killMe
{
	bKillMe = true;
}


//------------------------------------------------------------------------
// set whether ship is active or respawning
//------------------------------------------------------------------------
- (void) setActive:(bool)bAlive
{
	bActive = bAlive;
	
	if (bActive == false)
	{	// this net ship just died... explode it
		
		// adjust explode scale with a bit of randomness
		float baseBoomScale = scale * exploScale;
		float boomScale = baseBoomScale - (Rand_BetweenVals(0.0f, 0.2f) * baseBoomScale);
				
		GLAnim *pExplo = Explosion_Launch(EXPLODE_TYPE_LARGEWAVE, posX, posY, boomScale, sortDepth+1);
		[pExplo setVelocity:velocity.x andY:velocity.y andDamp:0.99f];
		
		// play big explosion sound
		Sound_Launch(SOUNDFX_EXPLODE, 1.0f);

		// return my leftover parts, and then re-setup for when I respawn
		[self resetShip];
		
		if (killedByUserID == gPlayerShipMsgID)
		{	// this enemy ship was killed by the local player
			gNumOnlineEnemiesKilledThisLife++;
			
			// check achievements
			if (gNumOnlineEnemiesKilledThisLife == 3)
			{
				[SocialNetworkingManager unlockAchievement:_3_KILL_STREAK];
			}
			else if (gNumOnlineEnemiesKilledThisLife == 6)
			{
				[SocialNetworkingManager unlockAchievement:_6_KILL_STREAK];
			}
			else if (gNumOnlineEnemiesKilledThisLife == 10)
			{
				[SocialNetworkingManager unlockAchievement:_10_KILL_STREAK];
			}
			else if (gNumOnlineEnemiesKilledThisLife == 20)
			{
				[SocialNetworkingManager unlockAchievement:_20_KILL_STREAK];
			}
			
			if (userID == gpPlayerSprite->revengeUserID)
			{	// player's next kill was the enemy that last killed the player
				[SocialNetworkingManager unlockAchievement:REVENGE];
			}
			gpPlayerSprite->revengeUserID = 0;
		}
	}
	else
	{	// just respawned... reset interpolation
		shieldStrength = maxShieldStrength;	// force shields to active on respawn
		shieldDownTimer = 0.0f;
		
		velocity.x = 0.0f;
		velocity.y = 0.0f;
		rotationSpeed = 0.0f;
		
		spawnScale = 0.01f;	// init for spawn effect
		GLAnim_LaunchAnim(&gSpawnAnim, posX, posY, Rand_BetweenVals(0.0f, 360.0f), sortDepth-1, scale*10.0f, false, true);
	}
}


//------------------------------------------------------------------------
// net ships do not respond to health... that's controlled from the 
// client controlling the ship
//------------------------------------------------------------------------
- (void) adjustHealth:(GLfloat)damage byID:(int)hitUserID
{
	if (shieldStrength > 0.0f)
	{	// set the shields to flash
		shieldDisplayRatio = 1.0f;
	}
	
	hitByUserID = hitUserID;
}


//------------------------------------------------------------------------
// set net ship's shield on or off
//------------------------------------------------------------------------
- (void) setShield:(bool)bOn howLong:(float)shieldTime
{
	if (bOn == true)
	{
		shieldStrength = maxShieldStrength;
	}
	else
	{
		shieldStrength = 0.0f;
		shieldDownTimer = shieldTime * MAX_FRAMERATE;
	}
}


//------------------------------------------------------------------------
// force the ship to a position (resets both display and final pos)
//------------------------------------------------------------------------
- (void) forcePos:(float)xPos andY:(float)yPos andRot:(float)rotPos
{
	posX = xPos;
	posY = yPos;
	rotation = rotPos;
}


//------------------------------------------------------------------------
// update the position/etc. of a net ship from a vector update packet
//------------------------------------------------------------------------
- (void) updateFromVectorMsg:(VectorUpdatePacket *)pMsg myPing:(int)localPing
{
	if (pMsg == nil)
	{
		return;
	}
	
	msgPos.x = pMsg.x;
	msgPos.y = pMsg.y;
	msgVelocity.x = pMsg.velocityX;
	msgVelocity.y = pMsg.velocityY;
	msgRotation = pMsg.rotation;
	msgRotSpeed = pMsg.angularVelocity;
	msgInputAngle = pMsg.inputAngle;
	msgInputMagnitude = pMsg.inputMagnitude;

	bNewPos = true;
		
	// calculate prediction time from ping values
	float pingTimeMS = (float)(pMsg.ping + localPing) / 2.0f;
	msgPingTime = pingTimeMS * MAX_FRAMERATE / 1000.0f;	// convert from ms to ticks
}


//------------------------------------------------------------------------
// update the position/etc. of a net ship from a position packet
//------------------------------------------------------------------------
- (void) updateFromPositionMsg:(PositionPacket *)pMsg
{
	if (pMsg == nil)
	{
		return;
	}
	
	msgPos.x = pMsg.x;
	msgPos.y = pMsg.y;
	msgVelocity.x = msgVelocity.y = 0.0f;
	msgRotSpeed = 0.0f;
	bNewPos = true;
}


//------------------------------------------------------------------------
// update the net ship
//------------------------------------------------------------------------
- (bool) updateSprite
{
	if (bKillMe == true)
	{	// get rid of this net ship
		if (bActive == true)
		{	// if ship was currently active, start warp out
			spawnScale = finalScale;
			GLAnim_LaunchAnim(&gSpawnAnim, posX, posY, Rand_BetweenVals(0.0f, 360.0f), sortDepth-1, scale*10.0f, false, true);
			bActive = false;
		}
		
		spawnScale -= gFrameTime * 0.1f;
		if (spawnScale <= 0.0f)
		{
			return true;
		}
		
		return false;
	}
	
	// handle name text
	if (pNameText == nil)
	{	// launch name text display
		char nameTxt[64];
		sprintf(nameTxt, "%s", [playername UTF8String]);
		pNameText = GLFont_LaunchText(posX, posY, 29, 0.1f, nameTxt);
		
		[pNameText setColorType:teamColor];
	}
	
	if (pChatIcon == nil)
	{	// launch the chat icon
		pChatIcon = GLChatIcon_Launch(29, 5.0f);
	}
	
	if (bActive == false)
	{	// ship is dead... don't display/collide/etc.
		bOffScreen = true;
		
		pNameText->bDrawMe = false;
		
		return false;
	}
	
	if (spawnScale < finalScale)
	{	// spawning effect
		spawnScale += gFrameTime * 0.1f;
		return false;
	}
	
	if (bNewPos == true)
	{	// got a vector update this frame...
		bNewPos = false;
		
		posX = msgPos.x;
		posY = msgPos.y;
		rotation = msgRotation;
		velocity = msgVelocity;
		rotationSpeed = msgRotSpeed;

		// set engines to new setting
		[self activateEngines:msgInputAngle distFromCenter:msgInputMagnitude];
	}

	// check for active parts
	ShipPartData *pCurPart = pPartsList;
	while (pCurPart != nil)
	{
		ShipPartTypeData *pTexData = ShipParts_GetTypeInfo(pCurPart->partNum);
		
		// check for attached effects
		switch (pTexData->partType)
		{
			case SHIP_PART_TYPE_ENGINE:
				if (pCurPart->activeAmount > 0.0f)
				{	// this part is active			
					CGPoint offsetPoint;
					
					// part is an engine... apply thrust from the engine's attach point
					GLPartsShip_GetAttachPoint(posX, posY, rotation, scale, pCurPart, &offsetPoint.x, &offsetPoint.y);
					
					GLfloat thrustAng = rotation + pCurPart->rotation;
					while (thrustAng > 360.0f)
					{
						thrustAng -= 360.0f;
					}
					
					if (pCurPart->health < (pCurPart->maxHealth / 2.0f))
					{	// part is damaged... randomly adjust thrust
						pCurPart->partData = Rand_BetweenVals(0.0f, pCurPart->activeAmount);
					}
					else
					{
						pCurPart->partData = pCurPart->activeAmount;
					}
					
					float thrustScale = scale * pCurPart->scale * SHIP_ENGINE_THRUST * pCurPart->partData;
					
					[self applyForceRot:offsetPoint.x yPos:offsetPoint.y angle:thrustAng forceSize:thrustScale isContinuous:true];
				}
				break;
				
			default:
				break;
		}
		
		// goto next part
		pCurPart = pCurPart->pNext;
	}

	// run physics
	float updateTime;
	if (msgPingTime >= 0.0f)
	{	// just got a vector update... use ping time to predict new position
		updateTime = msgPingTime;
		msgPingTime = -1.0f;
	}
	else
	{	// otherwise, update physics as normal, off frame time
		updateTime = gFrameTime;
	}
	
	posX += velocity.x * updateTime;
	posY += velocity.y * updateTime;
	
	rotation += rotationSpeed * updateTime;
	if (rotation < 0.0f)
	{
		rotation += 360.0f;
	}
	else if (rotation >= 360.0f)
	{
		rotation -= 360.0f;
	}
	
	if (SHIP_LINEAR_DAMPING > 0.0f)
	{	// dampen linear velocity
		velocity.x *= SHIP_LINEAR_DAMPING;
		velocity.y *= SHIP_LINEAR_DAMPING;
	}
	
	if (SHIP_ROTATIONAL_DAMPING > 0.0f)
	{	// dampen rotational velocity
		rotationSpeed *= SHIP_ROTATIONAL_DAMPING;
	}	
	
	// set screen zoom based on distance from player ship
	GLfloat dX = posX - gScreenWorldPos.x;
	GLfloat dY = posY - gScreenWorldPos.y;
	distFromPlayer = sqrt((dX*dX) + (dY*dY));
	
	// set camera distance
	Camera_UpdateTargetDist(distFromPlayer);
	
	// set on/off screen
	if (distFromPlayer > OFFSCREEN_DIST)
	{
		bOffScreen = true;
	}
	else
	{
		bOffScreen = false;
	}

	// update name text
	if (pNameText != nil)
	{
		pNameText->bDrawMe = !bOffScreen;
		
		if (bOffScreen == false)
		{	// adjust position
			CGPoint drawPos;
			drawPos.x = posX - gScreenWorldPos.x;
			drawPos.y = posY - gScreenWorldPos.y;
			Math_RotatePoint(drawPos.x, drawPos.y, -gfScreenRotation, &pNameText->startPosX, &pNameText->startPosY);
			
			//float textPosScale = (DISPLAY_SCALE_MAX - gfDisplayScale) + DISPLAY_SCALE_MIN;
			pNameText->startPosX += NAMETEXT_OFFSET_X; // * textPosScale;
			pNameText->startPosY += NAMETEXT_OFFSET_Y; // * textPosScale;

			pNameText->startPosX /= gfDisplayScale;
			pNameText->startPosY /= gfDisplayScale;

			if (pChatIcon != nil)
			{
				pChatIcon->bDrawMe = bChatting;
					
				pChatIcon->posX = pNameText->startPosX + CHAT_ICON_OFFSET_ADJUST_FROM_NAME_X;
				pChatIcon->posY = pNameText->startPosY + CHAT_ICON_OFFSET_ADJUST_FROM_NAME_Y;
			}
		}
		else
		{
			if (pChatIcon != nil)
			{
				pChatIcon->bDrawMe = false;
			}
		}
	}
		
	// adjust shield
	if (shieldDownTimer > 0.0f)
	{	// shield is down
		shieldDownTimer -= gFrameTime;
		if (shieldDownTimer <= 0.0f)
		{	// put back shield
			shieldStrength = maxShieldStrength;
		}
	}

	// adjust tilt scale
	float myRotSpeed;
	if (rotationSpeed < 0.0f)
	{
		myRotSpeed = -rotationSpeed;
	}
	else
	{
		myRotSpeed = rotationSpeed;
	}
	
	tiltScale = myRotSpeed / maxRotSpeed;
	tiltScale = 1.0f - (tiltScale * (1.0f - MAX_TILT_SCALE));
	
	return false;
}

	
-(void) dealloc
{
	if (playername) [playername release];
	if (pNameText != nil)
	{
		Sprite_RemoveDisplaySprite(pNameText);
		[pNameText release];		
	}
	
	if (pChatIcon != nil)
	{
		Sprite_RemoveDisplaySprite(pChatIcon);
		[pChatIcon release];
	}

	[super dealloc];
}


//-----------------------------------------------------------------------
// launch a net ship
//
// IN:	userID = net ID for this ship
//      name = playerName
//		posX/Y = world position to launch at
//		sortDepth = display depth
//		rotation = rotation to lunch at
//		scale = display scale
//-----------------------------------------------------------------------
+ (GLNetShip *) NetShip_LaunchWithUserID:(int)userID playerName:(NSString *)name x:(GLfloat)posX y:(GLfloat)posY sort:(int)sortDepth size:(GLfloat)scale rot:(GLfloat)rotation
{
	SpriteDisplayData *pSprite;
	
	pSprite = Sprite_NewDisplaySprite(sortDepth);
	if (pSprite != nil)
	{
		GLNetShip *pShip = (GLNetShip *)[GLNetShip createSprite:sortDepth];
		pSprite->pSprite = pShip;
		[pShip setScale:scale];
		[pShip forcePos:posX andY:posY andRot:rotation];
		
		pShip->userID = userID;
		pShip->playername = name;
		pShip->finalScale = scale;
        
		// mark this sprite as an enemy
		// NOTE: for team on-line, we need to mark the ship as friend or enemy, and may want to do this after
		//       the local player spawns (once we know the player's team)... we may need to do the marking of
		//       this ship in an external function....
		MarkedSprites_AddMarkedSprite(pShip, MARKEDSPRITE_TYPE_ENEMY);
        
		return pShip;
	}
	
	return(nil);
}


@end
