
#import <math.h>

#import "GLShip.h"
#import "spriteSystem.h"
#import "game.h"


@implementation GLShip


//------------------------------------------------------------------------
// ship specific init
//------------------------------------------------------------------------
- (GLSprite *)init:(int)depth
{
	// do base class init
	[super init:depth];
	
	maximumSpeed = 0.02f;
	curSpeed = 0.0f;
	acceleration = 0.0005f;
	
	turnSpeed = 3.0f;
	
	rotationTarget = -90.0f;
	speedTarget = 0.0f;
	
	bDoDamage = true;
	
	bUsePhysics = false;
	velocity.x = velocity.y = 0.0f;
	shipMass = 100.0f;
	shipRotInertia = 100.0f;
	
	bForcesApplied = false;
	bPrevForcesApplied = false;
	
	exploScale = 2.0f;
	
	return (self);
}


//------------------------------------------------------------------------
// set maximum flight speed
//------------------------------------------------------------------------
- (void) setMaxSpeed:(float)maxSpeed
{
	maximumSpeed = maxSpeed;
}


//------------------------------------------------------------------------
// set turn speed
//------------------------------------------------------------------------
- (void) setTurnSpeed:(float)rotSpeed
{
	turnSpeed = rotSpeed;
}


//------------------------------------------------------------------------
// set acceleration
//------------------------------------------------------------------------
- (void) setAcceleration:(float)accel
{
	acceleration = accel;
}


//------------------------------------------------------------------------
// set rotation target
//------------------------------------------------------------------------
- (void) setRotationTarget:(float)rotTarget
{
	rotationTarget = rotTarget;
	if (rotationTarget < 0.0f)
	{
		rotationTarget += 360.0f;
	}
	else if (rotationTarget > 360.0f)
	{
		rotationTarget -= 360.0f;
	}
}


//------------------------------------------------------------------------
// set speed target
//------------------------------------------------------------------------
- (void) setSpeedTarget:(float)spdTarget
{
	speedTarget = spdTarget;
	
	if (speedTarget < 0.0f)
	{
		speedTarget = 0.0f;
	}
	else if (speedTarget > maximumSpeed)
	{
		speedTarget = maximumSpeed;
	}
}


//------------------------------------------------------------------------
// adjust a ship's health by a damage amount
//------------------------------------------------------------------------
- (void) adjustHealth:(float)damage byID:(int)hitUserID
{
	if (shieldStrength > 0.0f)
	{	// this ship has shields
		shieldStrength -= damage;
		if (shieldStrength < 0.0f)
		{	// took down the shields
			damage = -shieldStrength;
		}
		else
		{
			damage = 0.0f;
			
			// set the shields to flash
			shieldDisplayRatio = 1.0f;
		}
	}
	
	health -= damage;
	hitByUserID = hitUserID;
}


//--------------------------------------------------------------------------
// apply a force at a point, with the force defined as a rotation and scale
//--------------------------------------------------------------------------
- (void) applyForceRot:(float)atX yPos:(float)atY angle:(float)rotAngle forceSize:(float)forceScale isContinuous:(bool)bContinuous
{
	CGPoint forceVect;
	forceVect.x = forceScale * -cos(rotAngle * DEG_TO_RAD);
	forceVect.y = forceScale * -sin(rotAngle * DEG_TO_RAD);

	CGPoint linAccel;
	float rotAccel;
	Physics_ApplyForce(shipMass, shipRotInertia, posX, posY, forceVect.x, forceVect.y, atX, atY, &linAccel.x, &linAccel.y, &rotAccel);
	
	if (bContinuous == false)
	{	// apply impulse
		velocity.x += linAccel.x;
		velocity.y += linAccel.y;
		rotationSpeed += rotAccel;
	}
	else
	{	// apply framerate adjusted impulse (for continuous thrust)
		velocity.x += linAccel.x * gFrameTime;
		velocity.y += linAccel.y * gFrameTime;
		rotationSpeed += rotAccel * gFrameTime;
	}
		
	bForcesApplied = true;
}


//--------------------------------------------------------------------------
// apply a force at a point, with the force defined as a 2D vector
//--------------------------------------------------------------------------
- (void) applyForceVector:(float)atX yPos:(float)atY forceX:(float)xForce forceY:(float)yForce
{
	CGPoint linAccel;
	float rotAccel;
	Physics_ApplyForce(shipMass, shipRotInertia, posX, posY, xForce, yForce, atX, atY, &linAccel.x, &linAccel.y, &rotAccel);
	
	velocity.x += linAccel.x;
	velocity.y += linAccel.y;
	rotationSpeed += rotAccel;
	
	bForcesApplied = true;
}


//------------------------------------------------------------------------
// update a ship
//------------------------------------------------------------------------
- (bool) updateSprite
{
	if (bDoDamage == true)
	{
		if (health <= 0.0f)
		{	// this object ran out of health... kill it with an explosion
			
			// adjust explode scale with a bit of randomness
			float baseBoomScale = scale * exploScale;
			float boomScale = baseBoomScale - (Rand_BetweenVals(0.0f, 0.2f) * baseBoomScale);
						
			GLAnim *pExplo = Explosion_Launch(EXPLODE_TYPE_LARGEWAVE, posX, posY, boomScale, sortDepth+1);
			[pExplo setVelocity:velocity.x andY:velocity.y andDamp:0.99f];
			
			// play big explosion sound
			Sound_Launch(SOUNDFX_EXPLODE, 1.0f);
			
			return true;
		}
	}
	
	if (bUsePhysics == false)
	{
		// adjust ship rotation towards target rotation
		if (rotation != rotationTarget)
		{
			float turnDist = rotationTarget - rotation;
		
			if (turnDist > 0.0f)
			{
				if ((turnDist < turnSpeed) || (turnDist > (360.0f-turnSpeed)))
				{
					rotation = rotationTarget;
				}
				else if (turnDist > 180.0f)
				{
					rotation -= turnSpeed * gFrameTime;
				}
				else
				{
					rotation += turnSpeed * gFrameTime;
				}
			}
			else
			{
				if ((turnDist > -turnSpeed) || (turnDist < (turnSpeed-360.0f)))
				{
					rotation = rotationTarget;
				}
				else if (turnDist < -180.0f)
				{
					rotation += turnSpeed * gFrameTime;
				}
				else
				{
					rotation -= turnSpeed * gFrameTime;
				}
			}
		
			if (rotation < 0.0f)
			{
				rotation += 360.0;
			}
			else if (rotation >= 360.0f)
			{
				rotation -= 360.0f;
			}
		}
	
		// adjust speed based on acceleration
		if (curSpeed < speedTarget)
		{	// increase speed up to target
			curSpeed += acceleration * gFrameTime;
			if (curSpeed > speedTarget)
			{
				curSpeed = speedTarget;
			}
		}
		else if (curSpeed > speedTarget)
		{	// decrease speed to target
			curSpeed -= acceleration * gFrameTime;
			if (curSpeed < speedTarget)
			{
				curSpeed = speedTarget;
			}
		}
		
		// fly the ship through the world if moving
		if (curSpeed > 0.0f)
		{
			velocity.x = curSpeed * -cos(rotation * DEG_TO_RAD);
			velocity.y = curSpeed * -sin(rotation * DEG_TO_RAD);
			[self adjustPos:(velocity.x * gFrameTime) andY:(velocity.y * gFrameTime)];
		}
	}
	else
	{	// use physics
		posX += velocity.x * gFrameTime;
		posY += velocity.y * gFrameTime;
		
		rotation += rotationSpeed * gFrameTime;
		if (rotation < 0.0f)
		{
			rotation += 360.0f;
		}
		else if (rotation >= 360.0f)
		{
			rotation -= 360.0f;
		}
		
		if (SHIP_LINEAR_DAMPING > 0.0f)
		{	// dampen linear velocity
			velocity.x *= SHIP_LINEAR_DAMPING;
			velocity.y *= SHIP_LINEAR_DAMPING;
		}
		
		if (SHIP_ROTATIONAL_DAMPING > 0.0f)
		{	// dampen rotational velocity
			rotationSpeed *= SHIP_ROTATIONAL_DAMPING;
		}		
	}

	return false;
}


@end
