
#import <math.h>

#import "GLSprite.h"
#import "spriteSystem.h"
#import "game.h"
#import "markedSprites.h"


@implementation GLSprite


//------------------------------------------------------------------------
// init a displable sprite object
//
// IN:	display depth of the sprite
//
// RETURNS:	ptr to the sprite object
//------------------------------------------------------------------------
- (GLSprite *)init:(int)depth
{
	if (self = [super init])
	{
		spriteTexture = 0;

		objectType = OBJECT_TYPE_NONE;
		
		baseSize = 0.0f;
		collideSize = 0.0f;
		
		// set default settings
		posX = 0.0f;
		posY = 0.0f;
		sortDepth = depth;
		rotation = 0.0f;
		rotationSpeed = 0.0f;
		scale = 1.0f;
		
		scrollScale = -1.0f;	// default to 1:1 scrolling
				
		bTimeOut = false;

		// anim defaults
		frameTimer = 0.0f;
		ticksPerFrame = 0.0f;
		bLoop = false;
		bRemoveAtEndOfAnim = false;
		
		bDoDamage = false;
		health = 10.0f;
		hitByUserID = -1;
		
		bDrawMe = true;
		
		bWorldSprite = true;
		
		markedType = MARKEDSPRITE_TYPE_NONE;
		
		distFromPlayer = -1.0f;
		
		bActive = true;
		
		userID = -1000;
		
		teamColor = COLOR_NONE;
		
		return self;
	}
	
	return nil;
}


//------------------------------------------------------------------------
// init a displable sprite object
//
// IN:	display depth of the sprite
//
// RETURNS:	ptr to the sprite object
//------------------------------------------------------------------------
+ (GLSprite *)createSprite:(int)depth;
{
	return [[self alloc] init:depth];
}


//------------------------------------------------------------------------
// set the position of the sprite
//------------------------------------------------------------------------
- (void) setPos:(GLfloat)x andY:(GLfloat)y
{
	posX = x;
	posY = y;
}


//------------------------------------------------------------------------
// adjust the position of the sprite by some amount
//------------------------------------------------------------------------
- (void) adjustPos:(GLfloat)x andY:(GLfloat)y
{
	posX += x;
	posY += y;
}


//------------------------------------------------------------------------
// set the rotation of the sprite
//------------------------------------------------------------------------
- (void) setRotation:(GLfloat)angle
{
	rotation = angle;
}


//------------------------------------------------------------------------
// adjust the rotation of the sprite by some amount
//------------------------------------------------------------------------
- (void) adjustRotation:(GLfloat)angle
{
	rotation += angle;
}


//------------------------------------------------------------------------
// set the rotation speed of the sprite
//------------------------------------------------------------------------
- (void) setRotationSpeed:(GLfloat)angle
{
	rotationSpeed = angle;
}


//------------------------------------------------------------------------
// set the scale of the sprite
//------------------------------------------------------------------------
- (void) setScale:(GLfloat)siz
{
	scale = siz;
}


//------------------------------------------------------------------------
// adjust the scale of the sprite by some amount
//------------------------------------------------------------------------
- (void) adjustScale:(GLfloat)siz
{
	scale += siz;
}


//------------------------------------------------------------------------
// set the time out timer (this sprite will now time out)
//------------------------------------------------------------------------
- (void) setTimeOutTime:(GLfloat)timeLasts
{
	bTimeOut = true;
	timeOutTimer = timeLasts;
}


//------------------------------------------------------------------------
// set the animation settings
//------------------------------------------------------------------------
- (void) animSettings:(GLfloat)frameTime loop:(bool)bLoopMe remove:(bool)bRemove
{
	ticksPerFrame = frameTime;
	bLoop = bLoopMe;
	bRemoveAtEndOfAnim = bRemove;
}


//------------------------------------------------------------------------
// check basic collision (simple distance check)
//------------------------------------------------------------------------
- (bool) checkCollide:(float)x andY:(float)y dam:(float)damageAmt hitBy:(int)hitByUser shieldFlag:(bool*)pbShielded
{
	float l_collideSize = collideSize * scale;
	float distX = x - posX;
	float distY = y - posY;
	
	if (distX < 0) distX = -distX;
	if (distY < 0) distY = -distY;
	
	// 1st pass collision, simple box
	if ((distX <= l_collideSize) && (distY <= l_collideSize))
	{	// might have collided... do detailed collision
		float distSqr = (distX*distX) + (distY*distY);
		if (distSqr <= (l_collideSize * l_collideSize))
		{	// hit it!
			[self adjustHealth:damageAmt byID:hitByUser];
			
			if (pbShielded != nil)
			{
				*pbShielded = false;
			}
			
			return(true);
		}
	}
	
	return(false);
}
	

//------------------------------------------------------------------------
// check basic collision (simple distance check)
// NOTE: this does similar to the checkCollide (above) or a basic 
// sprite, but in the case of sub-classes that override the checkCollide 
// function (such as ships that sometimes collide with shields, etc), 
// this function can be used to do simple collisions (for powerups, etc.)
//------------------------------------------------------------------------
- (bool) checkCollideBasic:(float)x andY:(float)y
{
	float l_collideSize = collideSize * scale;
	float distX = x - posX;
	float distY = y - posY;
	
	if (distX < 0) distX = -distX;
	if (distY < 0) distY = -distY;
	
	// 1st pass collision, simple box
	if ((distX <= l_collideSize) && (distY <= l_collideSize))
	{	// might have collided... do detailed collision
		float distSqr = (distX*distX) + (distY*distY);
		if (distSqr <= (l_collideSize * l_collideSize))
		{	// hit it!			
			return(true);
		}
	}
	
	return(false);
}


//-----------------------------------------------------------------
// check collision against a circle
//
// RETURNS:	penetration distance (negative if no collide)
//-----------------------------------------------------------------
- (float) checkCollideWithCircle:(float)x andY:(float)y andRad:(float)radius dam:(float)damageAmt hitBy:(int)hitByUser doPush:(bool)bDoPush
{
	float distX = x - posX;
	float distY = y - posY;
	
	// 1st pass, distance collide with ship ship/shield size
	float collideDist = radius + (collideSize * scale);
	
	float distSqr = (distX*distX) + (distY*distY);
	if (distSqr >= (collideDist * collideDist))
	{	// no collision possible
		return -1.0f;
	}		
	
	float penetration = collideDist - sqrt(distSqr);
		
	[self adjustHealth:damageAmt byID:hitByUser];
		
	return penetration;
}

	
//------------------------------------------------------------------------
// adjust a ship's health by a damage amount
//------------------------------------------------------------------------
- (void) adjustHealth:(GLfloat)damage byID:(int)hitUserID
{
	health -= damage;
	hitByUserID = hitUserID;
}


//-----------------------------------------------------------------------
// set the objects's team color
//-----------------------------------------------------------------------
- (void) setTeamColor:(int)teamClr
{
	teamColor = teamClr;
}


//------------------------------------------------------------------------
// update the sprite
//------------------------------------------------------------------------
- (bool) updateSprite
{
	bool bRet = false;

	// check time out
	if (bTimeOut == true)
	{
		timeOutTimer -= gFrameTime;
		if (timeOutTimer < 0.0f)
		{
			return(true);
		}
	}
	
	// rotate the sprite
	rotation += rotationSpeed * gFrameTime;
	if (rotation > 360.0f)
	{
		rotation -= 360.0;
	}
	else if (rotation < 0.0f)
	{
		rotation += 360.0f;
	}
	
	return bRet;
}


-(int)vertCount {
    return 4;
}


-(void)setupProjectionAndGetPoint:(CGPoint*)point
{
	if (bWorldSprite == true)
	{	// convert from world pos to screen pos
		point->x = posX - gScreenWorldPos.x;
		point->y = posY - gScreenWorldPos.y;
		
		if (scrollScale > 0.0f)
		{	// adjust for scroll ratio
			point->x *= scrollScale;
			point->y *= scrollScale;
		}
		
		if (gbDisplaySetting == false)
		{	// was previously drawing non-world sprites not affected by display scale
			CGPoint orthoVals;
			orthoVals.x = ORTHO_X * gfDisplayScale;
			orthoVals.y = ORTHO_Y * gfDisplayScale;
			
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrthof(-orthoVals.x, orthoVals.x, -orthoVals.y, orthoVals.y, -1.0f, 1.0f);
			glMatrixMode(GL_MODELVIEW);
			
			gbDisplaySetting = true;
		}
	}
	else
	{	// draw at raw screen position
		point->x = posX;
		point->y = posY;
		
		if (gbDisplaySetting == true)
		{	// was previously drawing world sprites... switch to screen sprites not affected by display scale
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrthof(-ORTHO_X, ORTHO_X, -ORTHO_Y, ORTHO_Y, -1.0f, 1.0f);
			glMatrixMode(GL_MODELVIEW);
			
			gbDisplaySetting = false;
		}
	}
}


//-------------------------------------------------------
// draw the sprite
//-------------------------------------------------------
- (void) drawSprite
{
	if (bDrawMe == false)
	{
		return;
	}
	
	CGPoint screenPos;
	[self setupProjectionAndGetPoint:&screenPos];
	
	//Bind this texture.
	glBindTexture(GL_TEXTURE_2D, spriteTexture);

	//Push the matrix so we can keep it as it was previously.
	glPushMatrix();

	if (bWorldSprite == true)
	{	// set camera rotation
		glRotatef(gfScreenRotation, 0.0f, 0.0f, 1.0f);
	}
	
	//Translate the OpenGL context to the center of the sprite for rotation and scale.
	glTranslatef(screenPos.x, screenPos.y, 0.0f);

	//Apply the rotation over the Z axis.
	glRotatef(rotation - 90.0f, 0.0f, 0.0f, 1.0f);

	// Scale relative to pixel size.
	glScalef(scale, scale, 1.0f);

	//Finally draw the arrays.
	glDrawArrays(GL_TRIANGLE_STRIP, 0, [self vertCount]);

	//Restore the model view matrix to prevent contamination.
	glPopMatrix();
}


-(void) dealloc
{
	if (markedType > MARKEDSPRITE_TYPE_NONE)
	{	// this sprite was marked... remove it from the marked list
		MarkedSprites_RemoveMarkedSprite(self);
	}
	
	[super dealloc];
}

@end
